@@ -1,4 +1,3 @@
-#!/usr/bin/python
 # -*- coding: utf-8 -*-
 #
 # Copyright (c) 2016 Red Hat, Inc.
@@ -19,7 +18,9 @@
 # along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
 #
 
+import collections
 import inspect
+import os
 import time
 
 from abc import ABCMeta, abstractmethod
@@ -29,7 +30,6 @@ from enum import Enum
 
 try:
     import ovirtsdk4 as sdk
-    import ovirtsdk4.types as otypes
     import ovirtsdk4.version as sdk_version
     HAS_SDK = LooseVersion(sdk_version.VERSION) >= LooseVersion('4.0.0')
 except ImportError:
@@ -52,41 +52,73 @@ def check_sdk(module):
         )
 
 
-def get_dict_of_struct(struct):
+def get_dict_of_struct(struct, connection=None, fetch_nested=False, attributes=None):
     """
     Convert SDK Struct type into dictionary.
     """
+    res = {}
+
     def remove_underscore(val):
         if val.startswith('_'):
             val = val[1:]
             remove_underscore(val)
         return val
 
-    res = {}
+    def convert_value(value):
+        nested = False
+
+        if isinstance(value, sdk.Struct):
+            return get_dict_of_struct(value)
+        elif isinstance(value, Enum) or isinstance(value, datetime):
+            return str(value)
+        elif isinstance(value, list) or isinstance(value, sdk.List):
+            if isinstance(value, sdk.List) and fetch_nested and value.href:
+                try:
+                    value = connection.follow_link(value)
+                    nested = True
+                except sdk.Error:
+                    value = []
+
+            ret = []
+            for i in value:
+                if isinstance(i, sdk.Struct):
+                    if not nested:
+                        ret.append(get_dict_of_struct(i))
+                    else:
+                        nested_obj = dict(
+                            (attr, convert_value(getattr(i, attr)))
+                            for attr in attributes if getattr(i, attr, None)
+                        )
+                        nested_obj['id'] = getattr(i, 'id', None),
+                        ret.append(nested_obj)
+                elif isinstance(i, Enum):
+                    ret.append(str(i))
+                else:
+                    ret.append(i)
+            return ret
+        else:
+            return value
+
     if struct is not None:
         for key, value in struct.__dict__.items():
-            key = remove_underscore(key)
             if value is None:
                 continue
-            elif isinstance(value, sdk.Struct):
-                res[key] = get_dict_of_struct(value)
-            elif isinstance(value, Enum) or isinstance(value, datetime):
-                res[key] = str(value)
-            elif isinstance(value, list):
-                res[key] = []
-                for i in value:
-                    if isinstance(i, sdk.Struct):
-                        res[key].append(get_dict_of_struct(i))
-                    elif isinstance(i, Enum):
-                        res[key].append(str(i))
-                    else:
-                        res[key].append(i)
-            else:
-                res[key] = value
+
+            key = remove_underscore(key)
+            res[key] = convert_value(value)
 
     return res
 
 
+def engine_version(connection):
+    """
+    Return string representation of oVirt engine version.
+    """
+    engine_api = connection.system_service().get()
+    engine_version = engine_api.product_info.version
+    return '%s.%s' % (engine_version.major, engine_version.minor)
+
+
 def create_connection(auth):
     """
     Create a connection to Python SDK, from task `auth` parameter.
@@ -113,6 +145,7 @@ def create_connection(auth):
         kerberos=auth.get('kerberos', None),
     )
 
+
 def convert_to_bytes(param):
     """
     This method convert units to bytes, which follow IEC standard.
@@ -166,7 +199,7 @@ def get_link_name(connection, link):
         return None
 
 
-def equal(param1, param2):
+def equal(param1, param2, ignore_case=False):
     """
     Compare two parameters and return if they are equal.
     This parameter doesn't run equal operation if first parameter is None.
@@ -178,6 +211,8 @@ def equal(param1, param2):
     :return: True if parameters are equal or first parameter is None, otherwise False
     """
     if param1 is not None:
+        if ignore_case:
+            return param1.lower() == param2.lower()
         return param1 == param2
     return True
 
@@ -196,7 +231,7 @@ def search_by_attributes(service, **kwargs):
     else:
         res = [
             e for e in service.list() if len([
-                 k for k, v in kwargs.items() if getattr(e, k, None) == v
+                k for k, v in kwargs.items() if getattr(e, k, None) == v
             ]) == len(kwargs)
         ]
 
@@ -233,6 +268,33 @@ def search_by_name(service, name, **kwargs):
     return res[0]
 
 
+def get_entity(service):
+    """
+    Ignore SDK Error in case of getting an entity from service.
+    """
+    entity = None
+    try:
+        entity = service.get()
+    except sdk.Error:
+        # We can get here 404, we should ignore it, in case
+        # of removing entity for example.
+        pass
+    return entity
+
+
+def get_id_by_name(service, name, raise_error=True, ignore_case=False):
+    """
+    Search an entity ID by it's name.
+    """
+    entity = search_by_name(service, name)
+
+    if entity is not None:
+        return entity.id
+
+    if raise_error:
+        raise Exception("Entity '%s' was not found." % name)
+
+
 def wait(
     service,
     condition,
@@ -256,15 +318,60 @@ def wait(
         start = time.time()
         while time.time() < start + timeout:
             # Exit if the condition of entity is valid:
-            entity = service.get()
+            entity = get_entity(service)
             if condition(entity):
                 return
             elif fail_condition(entity):
                 raise Exception("Error while waiting on result state of the entity.")
 
-            # Sleep for `poll_interval` seconds if nor of the conditions apply:
+            # Sleep for `poll_interval` seconds if none of the conditions apply:
             time.sleep(float(poll_interval))
 
+        raise Exception("Timeout exceed while waiting on result state of the entity.")
+
+
+def __get_auth_dict():
+    OVIRT_URL = os.environ.get('OVIRT_URL')
+    OVIRT_USERNAME = os.environ.get('OVIRT_USERNAME')
+    OVIRT_PASSWORD = os.environ.get('OVIRT_PASSWORD')
+    OVIRT_TOKEN = os.environ.get('OVIRT_TOKEN')
+    OVIRT_CAFILE = os.environ.get('OVIRT_CAFILE')
+    OVIRT_INSECURE = OVIRT_CAFILE is None
+
+    env_vars = None
+    if OVIRT_URL and ((OVIRT_USERNAME and OVIRT_PASSWORD) or OVIRT_TOKEN):
+        env_vars = {
+            'url': OVIRT_URL,
+            'username': OVIRT_USERNAME,
+            'password': OVIRT_PASSWORD,
+            'insecure': OVIRT_INSECURE,
+            'token': OVIRT_TOKEN,
+            'ca_file': OVIRT_CAFILE,
+        }
+    if env_vars is not None:
+        auth = dict(default=env_vars, type='dict')
+    else:
+        auth = dict(required=True, type='dict')
+
+    return auth
+
+
+def ovirt_facts_full_argument_spec(**kwargs):
+    """
+    Extend parameters of facts module with parameters which are common to all
+    oVirt facts modules.
+
+    :param kwargs: kwargs to be extended
+    :return: extended dictionary with common parameters
+    """
+    spec = dict(
+        auth=__get_auth_dict(),
+        fetch_nested=dict(default=False, type='bool'),
+        nested_attributes=dict(type='list', default=list()),
+    )
+    spec.update(kwargs)
+    return spec
+
 
 def ovirt_full_argument_spec(**kwargs):
     """
@@ -274,10 +381,12 @@ def ovirt_full_argument_spec(**kwargs):
     :return: extended dictionary with common parameters
     """
     spec = dict(
-        auth=dict(required=True, type='dict'),
+        auth=__get_auth_dict(),
         timeout=dict(default=180, type='int'),
         wait=dict(default=True, type='bool'),
         poll_interval=dict(default=3, type='int'),
+        fetch_nested=dict(default=False, type='bool'),
+        nested_attributes=dict(type='list', default=list()),
     )
     spec.update(kwargs)
     return spec
@@ -291,6 +400,24 @@ def check_params(module):
         module.fail_json(msg='"name" or "id" is required')
 
 
+def engine_supported(connection, version):
+    return LooseVersion(engine_version(connection)) >= LooseVersion(version)
+
+
+def check_support(version, connection, module, params):
+    """
+    Check if parameters used by user are supported by oVirt Python SDK
+    and oVirt engine.
+    """
+    api_version = LooseVersion(engine_version(connection))
+    version = LooseVersion(version)
+    for param in params:
+        if module.params.get(param) is not None:
+            return LooseVersion(sdk_version.VERSION) >= version and api_version >= version
+
+    return True
+
+
 class BaseModule(object):
     """
     This is base class for oVirt modules. oVirt modules should inherit this
@@ -305,6 +432,7 @@ class BaseModule(object):
         self._module = module
         self._service = service
         self._changed = changed
+        self._diff = {'after': dict(), 'before': dict()}
 
     @property
     def changed(self):
@@ -328,6 +456,12 @@ class BaseModule(object):
         """
         pass
 
+    def param(self, name, default=None):
+        """
+        Return a module parameter specified by it's name.
+        """
+        return self._module.params.get(name, default)
+
     def update_check(self, entity):
         """
         This method handle checks whether the entity values are same as values
@@ -362,7 +496,24 @@ class BaseModule(object):
         """
         pass
 
-    def create(self, entity=None, result_state=None, fail_condition=lambda e: False, search_params=None, **kwargs):
+    def diff_update(self, after, update):
+        for k, v in update.items():
+            if isinstance(v, collections.Mapping):
+                after[k] = self.diff_update(after.get(k, dict()), v)
+            else:
+                after[k] = update[k]
+        return after
+
+
+    def create(
+        self,
+        entity=None,
+        result_state=None,
+        fail_condition=lambda e: False,
+        search_params=None,
+        update_params=None,
+        **kwargs
+    ):
         """
         Method which is called when state of the entity is 'present'. If user
         don't provide `entity` parameter the entity is searched using
@@ -378,6 +529,7 @@ class BaseModule(object):
         :param result_state: State which should entity has in order to finish task.
         :param fail_condition: Function which checks incorrect state of entity, if it returns `True` Exception is raised.
         :param search_params: Dictionary of parameters to be used for search.
+        :param update_params: The params which should be passed to update method.
         :param kwargs: Additional parameters passed when creating entity.
         :return: Dictionary with values returned by Ansible module.
         """
@@ -390,9 +542,29 @@ class BaseModule(object):
             # Entity exists, so update it:
             entity_service = self._service.service(entity.id)
             if not self.update_check(entity):
+                new_entity = self.build_entity()
                 if not self._module.check_mode:
-                    entity_service.update(self.build_entity())
+                    update_params = update_params or {}
+                    updated_entity = entity_service.update(
+                        new_entity,
+                        **update_params
+                    )
                     self.post_update(entity)
+
+                # Update diffs only if user specified --diff paramter,
+                # so we don't useless overload API:
+                if self._module._diff:
+                    before = get_dict_of_struct(
+                        entity,
+                        self._connection,
+                        fetch_nested=True,
+                        attributes=['name'],
+                    )
+                    after = before.copy()
+                    self.diff_update(after, get_dict_of_struct(new_entity))
+                    self._diff['before'] = before
+                    self._diff['after'] = after
+
                 self.changed = True
         else:
             # Entity don't exists, so create it:
@@ -422,7 +594,13 @@ class BaseModule(object):
         return {
             'changed': self.changed,
             'id': entity.id,
-            type(entity).__name__.lower(): get_dict_of_struct(entity),
+            type(entity).__name__.lower(): get_dict_of_struct(
+                struct=entity,
+                connection=self._connection,
+                fetch_nested=self._module.params.get('fetch_nested'),
+                attributes=self._module.params.get('nested_attributes'),
+            ),
+            'diff': self._diff,
         }
 
     def pre_remove(self, entity):
@@ -433,6 +611,12 @@ class BaseModule(object):
         """
         pass
 
+    def entity_name(self, entity):
+        return "{e_type} '{e_name}'".format(
+            e_type=type(entity).__name__.lower(),
+            e_name=getattr(entity, 'name', None),
+        )
+
     def remove(self, entity=None, search_params=None, **kwargs):
         """
         Method which is called when state of the entity is 'absent'. If user
@@ -461,7 +645,6 @@ class BaseModule(object):
         entity_service = self._service.service(entity.id)
         if not self._module.check_mode:
             entity_service.remove(**kwargs)
-
             wait(
                 service=entity_service,
                 condition=lambda entity: not entity,
@@ -474,7 +657,12 @@ class BaseModule(object):
         return {
             'changed': self.changed,
             'id': entity.id,
-            type(entity).__name__.lower(): get_dict_of_struct(entity),
+            type(entity).__name__.lower(): get_dict_of_struct(
+                struct=entity,
+                connection=self._connection,
+                fetch_nested=self._module.params.get('fetch_nested'),
+                attributes=self._module.params.get('nested_attributes'),
+            ),
         }
 
     def action(
@@ -544,7 +732,13 @@ class BaseModule(object):
         return {
             'changed': self.changed,
             'id': entity.id,
-            type(entity).__name__.lower(): get_dict_of_struct(entity),
+            type(entity).__name__.lower(): get_dict_of_struct(
+                struct=entity,
+                connection=self._connection,
+                fetch_nested=self._module.params.get('fetch_nested'),
+                attributes=self._module.params.get('nested_attributes'),
+            ),
+            'diff': self._diff,
         }
 
     def search_entity(self, search_params=None):
